﻿using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.GraphicsLibraryFramework;

namespace GraphicsManager.Objects;

public class RoundedButton : IRenderObject
{
	private RoundedRectangle _bounds, _inside;
	internal Label _label;
	public RoundedButton()
	{
		_bounds = new RoundedRectangle();
		_inside = new RoundedRectangle();
		_label = new Label();
		_bounds.MouseEnter += BoundsOnMouseEnter;
		_bounds.MouseLeave += BoundsOnMouseLeave;
		_bounds.Clicked += BoundsOnClicked;
	}

	private Task BoundsOnClicked(IRenderObject arg)
	{
		if (Clicked is not null) _ = Clicked.Invoke(this);
		return Task.CompletedTask;
	}

	private Task BoundsOnMouseLeave(IRenderObject arg)
	{
		if (MouseLeave is not null) _ = MouseLeave.Invoke(this);
		return Task.CompletedTask;
	}

	private Task BoundsOnMouseEnter(IRenderObject arg)
	{
		if (MouseEnter is not null) _ = MouseEnter.Invoke(this);
		return Task.CompletedTask;
	}

	public event Func<IRenderObject, Task>? WindowLoaded;
	public event Func<IRenderObject, Task>? MouseEnter;
	public event Func<IRenderObject, Task>? MouseLeave;
	public object? Tag { get; set; } = null;
	public int Radius { get => _bounds.Radius; set { _bounds.Radius = value; _inside.Radius = value; } }
	public int Border { get; set; } = 2;
	public int Smoothness { get => _bounds.Smoothness; set { _bounds.Smoothness = value; _inside.Smoothness = value; } }
	public ObjectAnchor Anchor { get => _bounds.Anchor; set { _bounds.Anchor = value; _inside.Anchor = value; _label.Anchor = value; } }
	public Font Font { get => _label.Font; set => _label.Font = value; }
	public string Text { get => _label.Text; set => _label.Text = value; }
	public bool Loaded { get; private set; } = false;
	public ContextMenu? ContextMenu { get => _bounds.ContextMenu; set => _bounds.ContextMenu = value; }
	public Vector2i Size
	{
		get
		{
			return _bounds.Size;
		}
		set
		{
			_bounds.Size = value;
			_inside.Size = new(value.X - (Border * 2), value.Y - (Border * 2));
			_label.Location = new( Location.X + Border + (((value.X - (Border * 2)) / 2) - (_label.Size.X / 2)), Location.Y + Border + (((value.Y - (Border * 2)) / 2) - (_label.Size.Y / 2)));
		}
	}
	public Vector2i Location { 
		get => _bounds.Location; 
		set
		{
			_bounds.Location = value;
			_inside.Location = new(value.X + Border, value.Y + Border);
			_label.Location = new( value.X  + Border + (((Size.X - (Border * 2)) / 2) - (_label.Size.X / 2)), value.Y + Border + (((Size.Y - (Border * 2)) / 2) - (_label.Size.Y / 2)));
		}
	}
	public Vector2 SizeAsFloat { get => _bounds.SizeAsFloat; }
	public Vector2 LocationAsFloat { get => _bounds.LocationAsFloat; }
	public Vector2i Distance { get => _bounds.Distance; internal set => _bounds.Distance = value;}
	public IParent? Parent { get; private set; } = null;
	public Window? Window { get; private set; } = null;
	public Color4 InsideColor { get => _inside.BackgroundColor; set => _inside.BackgroundColor = value; }
	public Color4 BorderColor { get => _bounds.BackgroundColor; set => _bounds.BackgroundColor = value; }
	public Color4 FontColor { get => _label.Color; set => _label.Color = value; }
	public bool Visible
	{
		get => _bounds.Visible;
		set 
		{ 
			_bounds.Visible = value;
			_inside.Visible = value;
			_label.Visible = value;
		}
	}
	public void Focus()
	{
		
	}
	public void UnFocus()
	{
		
	}
	public event Func<IRenderObject, Task>? Clicked;

	public void Clean()
	{
		_bounds.Clean();
		_inside.Clean();
		_label.Clean();
	}

	public void Draw()
	{
		if (!Visible || !Loaded) return;
		_bounds.Draw();
		_inside.Draw();
		_label.Draw();
	}

	public void LoadToParent(IParent Parent, Window Window)
	{
		if (Loaded) return;
		this.Parent = Parent;
		this.Window = Window;
		Loaded = true;
		_bounds.LoadToParent(Parent, Window);
		_inside.LoadToParent(Parent, Window);
		_label.LoadToParent(Parent, Window);
		Location = Location;
		if (WindowLoaded is not null) WindowLoaded.Invoke(this);
	}
}
