﻿using System.ComponentModel;
using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects;
using GraphicsManager.Objects.Core;
using OpenTK.Graphics.OpenGL4;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Desktop;

namespace GraphicsManager;

public class Window : NativeWindow , IParent
{
	protected override void Dispose(bool disposing)
	{
		for (int i = 0; i < Controls.Length; i++)
		{
			Controls[i].Clean();
		}
		Context.MakeNoneCurrent();
		base.Dispose(disposing);
	}

	internal IRenderObject? focused;

	internal ContextMenu? ActiveMenu { get; set; } = null;
	public IParent? Parent { get; } = null;
	public Vector2 LocationAsFloat { get; } = new Vector2(0f, 0f);
	public Window(NativeWindowSettings nativeWindowSettings) : base(nativeWindowSettings)
	{
		Context.MakeCurrent();
		Texture.TextureShader.Add(Context, new("RectangleTexture", true));
		Objects.Rectangle.DefaultShader.Add(Context, new("Rectangle", true));
		Label.DefaultTextShader.Add(Context, new("Label", true));
		Label._characters.Add(Context, new());
	}

	public Window() : base(new NativeWindowSettings())
	{
		
	}
	public Vector2i Position { get; } = new Vector2i(0, 0);
	public Color4 BackgroundColor { get; set; } = new Color4(0, 0, 0, 255);

	public ControlList Controls { get; } = new();
#region Cool Math Things
	public float[] RctToFloat(int x, int y, int Width, int Height, bool hastexture = false, float z = 0.0f)
	{
		if (hastexture)
		{
			return new float[20] {
				IntToFloat(x + Width), IntToFloat(y, true), z, 1.0f, 1.0f,// top r
                IntToFloat(x + Width), IntToFloat(y + Height, true), z, 1.0f, 0.0f,//b r
                IntToFloat(x), IntToFloat(y + Height, true), z, 0.0f, 0.0f,//bot l
                IntToFloat(x), IntToFloat(y, true), z, 0.0f, 1.0f// top l
            };
		}
		else
		{
			return new float[12] {
				IntToFloat(x + Width), IntToFloat(y, true), z,// top r
                IntToFloat(x + Width), IntToFloat(y + Height, true), z, //b r
                IntToFloat(x), IntToFloat(y + Height, true), z, //bot l
                IntToFloat(x), IntToFloat(y, true), z,// top l
            };
		}
	}

	public Vector3 PointToVector(float x, float y, float z = 0.0f)
	{
		return new Vector3(IntToFloat(x), IntToFloat(y, true), z);
	}

	public float IntToFloat(float p, bool Invert = false)
	{
		float Size = (Invert ? this.Size.Y : this.Size.X);
		double half = Math.Round((double)Size / (double)2, 1);
		double Per = Math.Round((double)1 / half, 15);
		if (p == half) return 0.0f;
		if (Invert)
		{
			if (p > half) return (float)(((double)(p - half) * Per) * -1);
			else return (float)(1 - (p * Per));
		}
		else
		{
			if (p > half) return (float)((double)(p - half) * Per);
			else return (float)((1 - (p * Per)) * -1);
		}
	}

	public float FloatToInt(float p, bool Invert = false)
	{
		float Size = (Invert ? this.Size.Y : this.Size.X);
		double half = Math.Round((double)Size / (double)2, 15);
		if (p == 0) return (int)half;
		if (Invert)
		{
			if (p < 0)
			{
				p *= -1;
				p++;
				return (float)(half * p);
			}
			else
			{
				return (float)(half - (p * half));
			}
		}
		else
		{
			if (p < 0)
			{
				p *= -1;
				p++;
				return (float)(Size - (half * p));
			}
			else
			{
				return (float)(p * half + half);
			}
		}
	}
#endregion

	public void ParentResize(ResizeEventArgs e)
	{
		if (e.Width == 0 && e.Height == 0) return;
		BlockDraw = true;
		base.OnResize(e);
		GL.Viewport(0, 0, e.Width, e.Height);
		for (int i = 0; i < Controls.Length; i++)
		{
			if (!Controls[i].Loaded) continue;
            bool top = (Controls[i].Anchor & ObjectAnchor.Top) == ObjectAnchor.Top;
            bool left = (Controls[i].Anchor & ObjectAnchor.Left) == ObjectAnchor.Left;
            bool right = (Controls[i].Anchor & ObjectAnchor.Right) == ObjectAnchor.Right;
            bool bottom = (Controls[i].Anchor & ObjectAnchor.Bottom) == ObjectAnchor.Bottom;
            if (!top && !bottom) { Controls[i].Anchor |= ObjectAnchor.Top; top = true; }
            if (!left && !right) { Controls[i].Anchor |= ObjectAnchor.Left; left = true; }
            int lx = (left ? Controls[i].Location.X : Size.X - Controls[i].Distance.X - Controls[i].Size.X);
            int ly = (top ? Controls[i].Location.Y : Size.Y - Controls[i].Distance.Y - Controls[i].Size.Y);
            int sy = (bottom ? Size.Y - Controls[i].Distance.Y - ly : Controls[i].Size.Y);
            int sx = (right ? Size.X - Controls[i].Distance.X - lx : Controls[i].Size.X);
            Controls[i].Size = new(sx, sy);
            Controls[i].Location = new(lx, ly);
            if (Controls[i] is IParent parent)
            {
            	parent.ParentResize(e);
            }
		}
		DrawFrame();
		BlockDraw = false;
	}

	private int frame = 0;

	public void ReportSizeUpdate(IRenderObject Control)
	{
		
	}

	protected override void OnResize(ResizeEventArgs e)
	{
		ParentResize(e);
	}

	protected override void OnClosing(CancelEventArgs e)
	{
		GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		GL.BindVertexArray(0);
		GL.UseProgram(0);
		for (int i = 0; i < Controls.Length; i++)
		{
			Controls[i].Clean();
		}
		base.OnClosing(e);
	}
	public Rendertype Rendertype { get; set; } = Rendertype.ControlUpdates;

	internal bool CanControleUpdate { get; private set; } = true;

	public int FPS { get; set; } = 0;
	public event Func<Window, Task>? WindowLoaded;
	public void StartRender()
	{
		Context.MakeCurrent();
		initthread = Thread.CurrentThread.ManagedThreadId;
		ProcessEvents();
		DrawFrame();
		if (WindowLoaded is not null) WindowLoaded.Invoke(this);
		while (Exists && IsVisible && !IsExiting)
		{
			ProcessEvents();
			bool u = (Rendertype & Rendertype.ControlUpdates) == Rendertype.ControlUpdates;
			if (!u) DrawFrame();
			if (invokes.Any())
			{
				for (int i = 0; i < invokes.Count; i++) invokes.Dequeue().Invoke();
			}
			Thread.Sleep(8);
		}
	}

	public bool BlockDraw { get; set; } = false;

	public void TryDraw()
	{
		if (!BlockDraw) DrawFrame();
	}
	
	private int initthread = 0;

	public bool InvokeRequired
	{
		get
		{
			return initthread != Thread.CurrentThread.ManagedThreadId;
		}
	}

	private Queue<Action> invokes = new();
	
	public void Invoke(Action A)
	{
		invokes.Enqueue(A);
	}

	public void DrawFrame()
	{
		Context.MakeCurrent();
		frame++;
		Console.WriteLine($"Drawing Frame: {frame}");
		GL.ClearColor(BackgroundColor.R, BackgroundColor.G, BackgroundColor.B, (BackgroundColor.A * -1) + 1);
		IEnumerable<IRenderObject> needload = Controls.Where(a => a.Loaded == false);
		
		if (needload.Any())
		{
			BlockDraw = true;
			foreach (IRenderObject obj in needload)
			{
				obj.LoadToParent(this, this);
			}
			BlockDraw = false;
		}
		GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
		for (int i = 0; i < Controls.Length; i++)
		{
			if (Controls[i].Loaded) Controls[i].Draw();
		}
		Context.SwapBuffers();
		
	}
}